#include "Entity3d.h"

int Entity3d::__lastid = 0;

Entity3d::Entity3d(void)
{
	rendered_type = RENDERED_NONE;

	Connectivity.ConnectionsValue = 0;

	xpos = ypos = 0.0f;
	width = length = height = 1.0f;

	id = __lastid++;
}

Entity3d::~Entity3d(void)
{
}

void Entity3d::Initialize(float w, float l, float h)
{
	width = ABS(w);
	length = ABS(l);
	height = ABS(h);
	rendered_type = RENDERED_CUBE;
	Connectivity.ConnectionsValue = 0;
}


//connection width also is the width/length of the central box
//everything else is based off of this. Tweak as needed.
#define CONNECTION_WIDTH		0.35f

//Connection length must adjust accordingly, to fit within the grid square.
#define CONNECTION_LENGTH		((GRID_SIZE - CONNECTION_WIDTH)/2)

//the translation must be performed in terms of CONNECTION_LENGTH.
//Here we translate 1/2 the length of the connection (achieving the edge)
//and then we move it a further 1/2 of the connection width (moving it out of the central box).
//kind of disgusting. But this is the precise translation needed.
#define TRANSLATE_CONNECTION	(0.5f + ((CONNECTION_WIDTH/2) / CONNECTION_LENGTH))


//#define CONNECTION_HEIGHT	1.0f

void Entity3d::Render(ConnectionStruct externalConnections, ConnectionStruct adjacentTunnels, int targetTexture)
{
	//position in correct location
	glPushMatrix();
//	height = CONNECTION_HEIGHT;

	//center
	glScalef(width, length, height);
	glTranslatef(xpos/width, ypos/length, 0);

	//big box
	if(Connectivity.ConnectionsValue == 0)
	{
		glPushMatrix();
	//	glScalef(1, 1, 2);
		DrawTopAndBottom(targetTexture);

		//cap sides not in use:
		if(!externalConnections.west)
			DrawSideFace(WEST, targetTexture);
		else if(adjacentTunnels.west)
		{
			//draw two extra walls
			glPushMatrix();
				glScalef(1.0f, CONNECTION_LENGTH, 1.0f);

				glTranslatef(0, -TRANSLATE_CONNECTION, 0);
				DrawSideFace(WEST, targetTexture);

				glTranslatef(0, 2*TRANSLATE_CONNECTION, 0);
				DrawSideFace(WEST, targetTexture);
			glPopMatrix();
		}

		if(!externalConnections.east)
			DrawSideFace(EAST, targetTexture);
		else if(adjacentTunnels.east)
		{
			//draw two extra walls
			glPushMatrix();
				glScalef(1.0f, CONNECTION_LENGTH, 1.0f);

				glTranslatef(0, -TRANSLATE_CONNECTION, 0);
				DrawSideFace(EAST, targetTexture);

				glTranslatef(0, 2*TRANSLATE_CONNECTION, 0);
				DrawSideFace(EAST, targetTexture);
			glPopMatrix();
		}

		if(!externalConnections.north)
			DrawSideFace(NORTH, targetTexture);
		else if(adjacentTunnels.north)
		{
			//draw two extra walls
			glPushMatrix();
				glScalef(CONNECTION_LENGTH, 1.0f, 1.0f);

				glTranslatef(-TRANSLATE_CONNECTION, 0, 0);
				DrawSideFace(NORTH, targetTexture);

				glTranslatef(2*TRANSLATE_CONNECTION, 0, 0);
				DrawSideFace(NORTH, targetTexture);
			glPopMatrix();
		}

		if(!externalConnections.south)
			DrawSideFace(SOUTH, targetTexture);
		else if(adjacentTunnels.south)
		{
			//draw two extra walls
			glPushMatrix();
				glScalef(CONNECTION_LENGTH, 1.0f, 1.0f);

				glTranslatef(-TRANSLATE_CONNECTION, 0, 0);
				DrawSideFace(SOUTH, targetTexture);

				glTranslatef(2*TRANSLATE_CONNECTION, 0, 0);
				DrawSideFace(SOUTH, targetTexture);
			glPopMatrix();
		}

		glPopMatrix();
	}
	else
	{
		//central box
		glPushMatrix();
		glScalef(CONNECTION_WIDTH, CONNECTION_WIDTH, 1.0f);

		DrawTopAndBottom(targetTexture);

		//cap sides not in use:
		if(!Connectivity.Connections.west)
			DrawSideFace(WEST, targetTexture);

		if(!Connectivity.Connections.east)
			DrawSideFace(EAST, targetTexture);

		if(!Connectivity.Connections.north)
			DrawSideFace(NORTH, targetTexture);

		if(!Connectivity.Connections.south)
			DrawSideFace(SOUTH, targetTexture);

		glPopMatrix();

		
		if(Connectivity.Connections.east)
		{
			glPushMatrix();
			glScalef(CONNECTION_LENGTH, CONNECTION_WIDTH, 1.0f);
			glTranslatef(TRANSLATE_CONNECTION, 0, 0);
			DrawTopAndBottom(targetTexture);
			DrawSideFace(NORTH, targetTexture);
			DrawSideFace(SOUTH, targetTexture);

			if(!externalConnections.east)
				DrawSideFace(EAST, targetTexture);

			glPopMatrix();
		}

		if(Connectivity.Connections.west)
		{
			glPushMatrix();
			glScalef(CONNECTION_LENGTH, CONNECTION_WIDTH, 1.0f);
			glTranslatef(-(TRANSLATE_CONNECTION), 0, 0);
			DrawTopAndBottom(targetTexture);
			DrawSideFace(NORTH, targetTexture);
			DrawSideFace(SOUTH, targetTexture);

			if(!externalConnections.west)
				DrawSideFace(WEST, targetTexture);

			glPopMatrix();
		}

		if(Connectivity.Connections.north)
		{
			glPushMatrix();
			glScalef(CONNECTION_WIDTH, CONNECTION_LENGTH, 1.0f);
			glTranslatef(0, TRANSLATE_CONNECTION, 0);
			DrawTopAndBottom(targetTexture);

			DrawSideFace(EAST, targetTexture);
			DrawSideFace(WEST, targetTexture);

			if(!externalConnections.north)
				DrawSideFace(NORTH, targetTexture);

			glPopMatrix();
		}

		if(Connectivity.Connections.south)
		{
			glPushMatrix();
			glScalef(CONNECTION_WIDTH, CONNECTION_LENGTH, 1.0f);
			glTranslatef(0, -(TRANSLATE_CONNECTION), 0);
			DrawTopAndBottom(targetTexture);
			DrawSideFace(EAST, targetTexture);
			DrawSideFace(WEST, targetTexture);

			if(!externalConnections.south)
				DrawSideFace(SOUTH, targetTexture);

			glPopMatrix();
		}
	}

	glPopMatrix();
}

void Entity3d::DrawFace(float yangle, float zangle)
{
	glPushMatrix();
	
	glRotatef(yangle, 0, 1.0f, 0); 
	glRotatef(zangle, 1.0f, 0, 0);

	glTranslatef(0,0,+0.5);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(-0.5f,	-0.5f, 0.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(0.5f, -0.5f, 0.0f);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(0.5f, 0.5f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(-0.5f, 0.5f, 0.0f);
	glEnd();

	glPopMatrix();
}

void Entity3d::DrawTopAndBottom(int targetTexture)
{
	if(targetTexture == 1)
	{
		DrawFace(0);
	}

	if(targetTexture == 2)
	{
		DrawFace(180);
	}
}

void Entity3d::DrawSideFace(Direction dir, int targetTexture)
{
	if(targetTexture == 0)
	{
		switch(dir)
		{
		case WEST:
			DrawFace(90, 180);
			break;
		case EAST:
			DrawFace(-90, 180);
			break;
		case NORTH:
			DrawFace(-90, -90);
			break;
		case SOUTH:
			DrawFace(90, 90);
			break;
		}
	}
}

bool Entity3d::ConstrainToIndoors(Vector3d &originalPoint, const Vector3d &delta, ConnectionStruct exits, ConnectionStruct adjacentTunnels)
{
	//this far away from walls.
	float margin = 0.1f;

	LOG("Exits: %s %s %s %s", exits.north ? "north" : "", exits.east ? "east" : "", exits.south ? "south" : "", exits.west ? "west" : "");

	bool was_inside = true;
	Vector3d newpoint = originalPoint + delta;
	float x_min, y_min, x_max, y_max;

	//first, collide against internal structure.
	if(Connectivity.ConnectionsValue != 0)
	{
		x_min = (xpos-CONNECTION_WIDTH/2) + margin;
		y_min = (ypos-CONNECTION_WIDTH/2) + margin;
		x_max = (xpos+CONNECTION_WIDTH/2) - margin;
		y_max = (ypos+CONNECTION_WIDTH/2) - margin;
		if(newpoint.x < x_min)
		{
			if(!Connectivity.Connections.west)
			{
				newpoint.x = x_min;
				was_inside = false;
			}
			else
			{
				//we are inside a corridor
				//restrict north/south in this section
				if((originalPoint.y >= y_min) && (newpoint.y < y_min))
				{
					newpoint.y = y_min;
					was_inside = false;
				}
				if((originalPoint.y <= y_max) && (newpoint.y > y_max))
				{
					newpoint.y = y_max;
					was_inside = false;
				}
			}
		}

		if(newpoint.y < y_min)
		{
			if(!Connectivity.Connections.south)
			{
				newpoint.y = y_min;
				was_inside = false;
			}
			else
			{
				//we are inside a corridor
				//restrict east/west in this section
				if((originalPoint.x >= x_min) && (newpoint.x < x_min))
				{
					newpoint.x = x_min;
					was_inside = false;
				}
				if((originalPoint.x <= x_max) && (newpoint.x > x_max))
				{
					newpoint.x = x_max;
					was_inside = false;
				}
			}

		}
		if(newpoint.x > x_max)
		{
			if(!Connectivity.Connections.east)
			{
				newpoint.x = x_max;
				was_inside = false;
			}
			else
			{
				//we are inside a corridor
				//restrict north/south in this section
				if((originalPoint.y >= y_min) && (newpoint.y < y_min))
				{
					newpoint.y = y_min;
					was_inside = false;
				}
				if((originalPoint.y <= y_max) && (newpoint.y > y_max))
				{
					newpoint.y = y_max;
					was_inside = false;
				}
			}
		}
		if(newpoint.y > y_max)
		{
			if(!Connectivity.Connections.north)
			{
				newpoint.y = y_max;
				was_inside = false;
			}
			else
			{
				//we are inside a corridor
				//restrict east/west in this section
				if((originalPoint.x >= x_min) && (newpoint.x < x_min))
				{
					newpoint.x = x_min;
					was_inside = false;
				}
				if((originalPoint.x <= x_max) && (newpoint.x > x_max))
				{
					newpoint.x = x_max;
					was_inside = false;
				}
			}
		}
	}

	//then, collide against the boundaries of the cell.
	x_min = xpos  - width/2 + margin;
	y_min = ypos  - length/2 + margin;
	x_max = xpos  + width/2 - margin;
	y_max = ypos  + length/2 - margin;
	if(newpoint.x < x_min)
	{
		if(!exits.west)
		{
			//no excuse, can't leave that way
			newpoint.x = x_min;
			was_inside = false;
		}
		else if((Connectivity.ConnectionsValue == 0) && (adjacentTunnels.west))
		{
			//additionally, constrain to only a portion of the exit.
			float ymin = (ypos-CONNECTION_WIDTH/2) + margin;
			float ymax = (ypos+CONNECTION_WIDTH/2) - margin;
			if(newpoint.y < ymin || newpoint.y > ymax)
			{
				newpoint.x = x_min;
				was_inside = false;
			}
		}
	}
	if(newpoint.y < y_min)
	{
		if(!exits.south)
		{
			newpoint.y = y_min;
			was_inside = false;
		}
		else if((Connectivity.ConnectionsValue == 0) && (adjacentTunnels.south))
		{
			//additionally, constrain to only a portion of the exit.
			float xmin = (xpos-CONNECTION_WIDTH/2) + margin;
			float xmax = (xpos+CONNECTION_WIDTH/2) - margin;
			if(newpoint.x < xmin || newpoint.x > xmax)
			{
				newpoint.y = y_min;
				was_inside = false;
			}
		}
	}
	if(newpoint.x > x_max)
	{
		if(!exits.east)
		{
			newpoint.x = x_max;
			was_inside = false;
		}
		else if((Connectivity.ConnectionsValue == 0) && (adjacentTunnels.east))
		{
			//additionally, constrain to only a portion of the exit.
			float ymin = (ypos-CONNECTION_WIDTH/2) + margin;
			float ymax = (ypos+CONNECTION_WIDTH/2) - margin;
			if(newpoint.y < ymin || newpoint.y > ymax)
			{
				newpoint.x = x_max;
				was_inside = false;
			}
			
		}
	}
	if(newpoint.y > y_max)
	{
		if(!exits.north)
		{
			newpoint.y = y_max;
			was_inside = false;
		}
		else if((Connectivity.ConnectionsValue == 0) && (adjacentTunnels.north))
		{
			//additionally, constrain to only a portion of the exit.
			float xmin = (xpos-CONNECTION_WIDTH/2) + margin;
			float xmax = (xpos+CONNECTION_WIDTH/2) - margin;
			if(newpoint.x < xmin || newpoint.x > xmax)
			{
				newpoint.y = y_max;
				was_inside = false;
			}
		}
	}

	originalPoint = newpoint;
	return was_inside;
}