#pragma once

#include <utils.h>
#include "Camera.h"
#include "Serializable.h"
#include "Character.h"
#include "ViewPort.h"
#include "GeneratedEntities.h"



#define MINIMUM_MAGNITUDE 1.0f

enum RenderMode
{
	RENDER_SOLID,
	RENDER_WIREFRAME
};

enum VisualMode
{
	VISUAL_TOPDOWN,
	VISUAL_CHARACTER
};

class World: public ViewPort, public Serializable
{
public:
	//singleton object
	static World* theWorld;

	VisualMode GetVisualMode();
	void SetVisualMode(VisualMode selected);
	void CycleVisualMode();

	int GetLayers(LayerInfo *infos);

	RenderMode GetRenderMode();
	void SetRenderMode(RenderMode selected);

	virtual void SetLightingAndCamera();

	void ResetCameras();

	virtual void Render();
	virtual void Update(bool force = false);

	virtual void PanMouse(float dx, float dy);
	virtual void ZoomMouse(float dx, float dy);

	void RemoveAllLayers();
	void AddLayer(Generator* newLayer);
	void RemoveLayer(int id = -1);
	void RemoveInputLayer(int selectedIndex);
	void SetInputLayer(int selectedVariable, int typeIndex);
	void SetWindowSize(int newWindowSize, int architectureIndex);
	int GetWindowSize(int architectureIndex);

	void NavigateInto(int selectedVariable);
	void NavigateUp();
	void SetSelectedLayer(int selectedID);

	void InverseChecked(bool checked);
	

	virtual bool DoFrame(int timeElapsed, int *keysDown);

	void GetWindowCenter(float &x, float &y);
	void GetCenterPosition(float &x, float &y);

	string GetLookDirection();

	void GetGridContainingPoint(float xpos, float ypos, float &x, float &y);
	void GetGridContainingPoint(const Vector3d &point, float &x, float &y);

	void BuildLighting();
	void CreateView();

	void SaveToXml(TiXmlElement *currentElement);
	void LoadFromXml(TiXmlElement *currentElement);

	bool ConstrainToIndoors(Vector3d &originalPoint, const Vector3d &delta);

	World(void);
	~World(void);

	GeneratorBase *GetCurrentLayer();
	bool IsCurrentLayer3d();

private:
	VisualMode m_visualmode;
	RenderMode m_rendermode;

	int GetIdIndex(int id);

	Camera camera;
	Character character;

	GeneratedEntities *m_architecture[MAX_RENDERABLE];
	int m_architectureCount;

	GeneratorBase	*m_currentLayer;
};